@@ -14,6 +14,11 @@ sub rpc {
 	return (@ret);
 }
 
+sub id {
+	my $self = shift;
+	return ${$self};
+}
+
 1;
 
 __END__
@@ -36,6 +41,10 @@ This module represents a thread ID object. It provides a handle to a thread.
 
 Send a message to a thread. The message items may contain any data type that can be serialized by L<Storable> (including coderefs).
 
+=head2 id
+
+Get the numerical id of this thread.
+
 =head2 monitor
 
 Monitor the thread. This will cause the calling thread to get a notification of the thread's end. In case of a natural death it returns C<('normal', @return_value)>, in case of an unnatural death it will contain C<('error', $exception)>.
