@@ -1,6 +1,7 @@
 #define PERL_NO_GET_CONTEXT
 #include "EXTERN.h"
 #include "perl.h"
+#define NO_XSLOCKS
 #include "XSUB.h"
 
 #include "queue.h"
@@ -27,23 +28,22 @@ static void S_message_set_sv(pTHX_ message* message, SV* value, enum node_type t
 
 #define message_set_sv(message, value, type) S_message_set_sv(aTHX_ message, value, type)
 
-static void S_message_store_value(pTHX_ message* message, SV* value) {
+void S_message_store_value(pTHX_ message* message, SV* value) {
 	dSP;
 	ENTER;
 	SAVETMPS;
 	sv_setiv(save_scalar(gv_fetchpv("Storable::Deparse", TRUE | GV_ADDMULTI, SVt_PV)), 1);
 	PUSHMARK(SP);
-	PUSHs(sv_2mortal(newRV_inc(value)));
+	XPUSHs(sv_2mortal(newRV_inc(value)));
 	PUTBACK;
 	call_pv("Storable::mstore", G_SCALAR);
 	SPAGAIN;
 	message_set_sv(message, POPs, STORABLE);
 	FREETMPS;
 	LEAVE;
+	PUTBACK;
 }
 
-#define message_store_value(message, value) S_message_store_value(aTHX_ message, value)
-
 void S_message_pull_stack(pTHX_ message* message) {
 	dSP; dMARK;
 	if (SP == MARK) {
@@ -53,37 +53,44 @@ void S_message_pull_stack(pTHX_ message* message) {
 			message_set_sv(message, *MARK, STRING);
 	}
 	else {
-		SV* list = sv_2mortal((SV*)av_make(SP - MARK + 1, MARK));
+		SV* list = sv_2mortal((SV*)av_make(SP - MARK, MARK + 1));
 		message_store_value(message, list);
 	}
 }
 
-#define message_pull_stack(message) STMT_START { PUTBACK; S_message_pull_stack(aTHX_ message); SPAGAIN; } STMT_END
-
-void S_message_push_stack(pTHX_ message* message) {
+SV* S_message_load_value(pTHX_ message* message) {
 	dSP;
 
+	sv_setiv(save_scalar(gv_fetchpv("Storable::Eval", TRUE | GV_ADDMULTI, SVt_PV)), 1);
+	PUSHMARK(SP);
+	XPUSHs(sv_2mortal(message_get_sv(message)));
+	PUTBACK;
+	call_pv("Storable::thaw", G_SCALAR);
+	SPAGAIN;
+	SV* ret = POPs;
+	PUTBACK;
+	return ret;
+}
+
+#define message_load_value(message) S_message_load_value(aTHX_ message)
+
+void S_message_push_stack(pTHX_ message* message, U32 context) {
+	dSP;
 	switch(message->type) {
 		case STRING:
-			PUSHs(sv_2mortal(message_get_sv(message)));
+			PUSHs(sv_2mortal(newRV_noinc(message_get_sv(message))));
 			break;
 		case STORABLE: {
-			ENTER;
-			sv_setiv(save_scalar(gv_fetchpv("Storable::Eval", TRUE | GV_ADDMULTI, SVt_PV)), 1);
-			PUSHMARK(SP);
-			XPUSHs(sv_2mortal(message_get_sv(message)));
-			PUTBACK;
-			call_pv("Storable::thaw", G_SCALAR);
+			AV* values = (AV*) SvRV(message_load_value(message));
 			SPAGAIN;
-			LEAVE;
-			AV* values = (AV*)SvRV(POPs);
 
-			if (GIMME_V == G_SCALAR) {
+			if (context == G_SCALAR) {
 				SV** ret = av_fetch(values, 0, FALSE);
 				PUSHs(ret ? *ret : &PL_sv_undef);
 			}
-			else if (GIMME_V == G_ARRAY) {
+			else if (context == G_ARRAY) {
 				UV count = av_len(values) + 1;
+				EXTEND(SP, count);
 				Copy(AvARRAY(values), SP + 1, count, SV*);
 				SP += count;
 			}
@@ -107,7 +114,7 @@ void S_message_clone(pTHX_ message* origin, message* clone) {
 			clone->string.ptr = savepvn(origin->string.ptr, origin->string.length);
 			break;
 		default:
-			warn("Unknown type in message\n");
+			Perl_die(aTHX, "Unknown type in message\n");
 	}
 }
 
@@ -122,8 +129,8 @@ void message_destroy(message* message) {
 			Safefree(message->string.ptr);
 			Zero(message, 1, message);
 			break;
-		default:
-			warn("Unknown type in message\n");
+//		default:
+//			Perl_die(aTHX, "Unknown type in message\n");
 	}
 }
 
